/* zexy-setup autogenerated setup-file
 * generated by "./makesource.sh"
 * !! DO NOT MANUALLY EDIT  !!
 */

/* forward declarations */
void setup_0x260x260x7e(void);         /*  0x260x260x7e.c      */
void setup_0x2e(void);                 /*  0x2e.c              */
void setup_0x3c0x7e(void);             /*  0x3c0x7e.c          */
void setup_0x3d0x3d0x7e(void);         /*  0x3d0x3d0x7e.c      */
void setup_0x3e0x7e(void);             /*  0x3e0x7e.c          */
void setup_0x7c0x7c0x7e(void);         /*  0x7c0x7c0x7e.c      */
void a2l_setup(void);                  /*  a2l.c               */
void absgn_tilde_setup(void);          /*  absgn~.c            */
void atof_setup(void);                 /*  atof.c              */
void atoi_setup(void);                 /*  atoi.c              */
void avg_tilde_setup(void);            /*  avg~.c              */
void blockmirror_tilde_setup(void);    /*  blockmirror~.c      */
void blockshuffle_tilde_setup(void);   /*  blockshuffle~.c     */
void blockswap_tilde_setup(void);      /*  blockswap~.c        */
void date_setup(void);                 /*  date.c              */
void demultiplex_setup(void);          /*  demultiplex.c       */
void demultiplex_tilde_setup(void);    /*  demultiplex~.c      */
void dfreq_tilde_setup(void);          /*  dfreq~.c            */
void dirac_tilde_setup(void);          /*  dirac~.c            */
void doublepole_tilde_setup(void);     /*  doublepole~.c       */
void drip_setup(void);                 /*  drip.c              */
void envrms_tilde_setup(void);         /*  envrms~.c           */
void envvar_setup(void);               /*  envvar.c           */
void fifop_setup(void);                /*  fifop.c             */
void freadln_setup(void);              /*  freadln.c           */
void fwriteln_setup(void);             /*  fwriteln.c          */
void glue_setup(void);                 /*  glue.c              */
void index_setup(void);                /*  index.c             */
void length_setup(void);               /*  length.c            */
void lifop_setup(void);                /*  lifop.c             */
void limiter_tilde_setup(void);        /*  limiter~.c          */
void list2int_setup(void);             /*  list2int.c          */
void list2lists_setup(void);           /*  list2lists.c        */
void list2symbol_setup(void);          /*  list2symbol.c       */
void lister_setup(void);               /*  lister.c            */
void listfind_setup(void);             /*  listfind.c          */
void liststorage_setup(void);          /*  liststorage.c       */
void longload_setup(void);             /*  longload.c          */
void lpt_setup(void);                  /*  lpt.c               */
void makesymbol_setup(void);           /*  makesymbol.c        */
void matchbox_setup(void);             /*  matchbox.c          */
void mavg_setup(void);                 /*  mavg.c              */
void minmax_setup(void);               /*  minmax.c            */
void msgfile_setup(void);              /*  msgfile.c           */
void multiline_tilde_setup(void);      /*  multiline~.c        */
void multiplex_setup(void);            /*  multiplex.c         */
void multiplex_tilde_setup(void);      /*  multiplex~.c        */
void multireceive_setup(void);         /*  multireceive.c      */
void niagara_setup(void);              /*  niagara.c           */
void noish_tilde_setup(void);          /*  noish~.c            */
void noisi_tilde_setup(void);          /*  noisi~.c            */
void operating_system_setup(void);     /*  operating_system.c  */
void zpack_setup(void);                /*  pack.c              */
void packel_setup(void);               /*  packel.c            */
void pack_tilde_setup(void);           /*  pack~.c             */
void pdf_tilde_setup(void);            /*  pdf~.c              */
void prime_setup(void);                /*  prime.c             */
void quantize_tilde_setup(void);       /*  quantize~.c         */
void rawprint_setup(void);             /*  rawprint.c          */
void regex_setup(void);                /*  regex.c             */
void relay_setup(void);                /*  relay.c             */
void repack_setup(void);               /*  repack.c            */
void repeat_setup(void);               /*  repeat.c            */
void route_tilde_setup(void);          /*  route~.c            */
void sfplay_setup(void);               /*  sfplay.c            */
void sfrecord_setup(void);             /*  sfrecord.c          */
void sgn_tilde_setup(void);            /*  sgn~.c              */
void sigzero_tilde_setup(void);        /*  sigzero~.c          */
void sleepgrain_setup(void);           /*  sleepgrain.c        */
void zexy_sort_setup(void);            /*  sort.c              */
void step_tilde_setup(void);           /*  step~.c             */
void strcmp_setup(void);               /*  strcmp.c            */
void sum_setup(void);                  /*  sum.c               */
void swap_tilde_setup(void);           /*  swap~.c             */
void symbol2list_setup(void);          /*  symbol2list.c       */
void tabdump_setup(void);              /*  tabdump.c           */
void tabminmax_setup(void);            /*  tabminmax.c         */
void tabread4_tilde_tilde_setup(void); /*  tabread4~~.c        */
void tabset_setup(void);               /*  tabset.c            */
void tavg_tilde_setup(void);           /*  tavg~.c             */
void time_setup(void);                 /*  time.c              */
void zunpack_setup(void);              /*  unpack.c            */
void unpack_tilde_setup(void);         /*  unpack~.c           */
void urn_setup(void);                  /*  urn.c               */
void wrap_setup(void);                 /*  wrap.c              */
void z_tilde_setup(void);              /*  z~.c                */

void z_zexy_setup(void)
{
  setup_0x260x260x7e();
  setup_0x2e();
  setup_0x3c0x7e();
  setup_0x3d0x3d0x7e();
  setup_0x3e0x7e();
  setup_0x7c0x7c0x7e();
  a2l_setup();
  absgn_tilde_setup();
  atof_setup();
  atoi_setup();
  avg_tilde_setup();
  blockmirror_tilde_setup();
  blockshuffle_tilde_setup();
  blockswap_tilde_setup();
  date_setup();
  demultiplex_setup();
  demultiplex_tilde_setup();
  dfreq_tilde_setup();
  dirac_tilde_setup();
  doublepole_tilde_setup();
  drip_setup();
  envrms_tilde_setup();
  envvar_setup();
  fifop_setup();
  freadln_setup();
  fwriteln_setup();
  glue_setup();
  index_setup();
  length_setup();
  lifop_setup();
  limiter_tilde_setup();
  list2int_setup();
  list2lists_setup();
  list2symbol_setup();
  lister_setup();
  listfind_setup();
  liststorage_setup();
  longload_setup();
  lpt_setup();
  makesymbol_setup();
  matchbox_setup();
  mavg_setup();
  minmax_setup();
  msgfile_setup();
  multiline_tilde_setup();
  multiplex_setup();
  multiplex_tilde_setup();
  multireceive_setup();
  niagara_setup();
  noish_tilde_setup();
  noisi_tilde_setup();
  operating_system_setup();
  zpack_setup();
  packel_setup();
  pack_tilde_setup();
  pdf_tilde_setup();
  prime_setup();
  quantize_tilde_setup();
  rawprint_setup();
  regex_setup();
  relay_setup();
  repack_setup();
  repeat_setup();
  route_tilde_setup();
  sfplay_setup();
  sfrecord_setup();
  sgn_tilde_setup();
  sigzero_tilde_setup();
  sleepgrain_setup();
  zexy_sort_setup();
  step_tilde_setup();
  strcmp_setup();
  sum_setup();
  swap_tilde_setup();
  symbol2list_setup();
  tabdump_setup();
  tabminmax_setup();
  tabread4_tilde_tilde_setup();
  tabset_setup();
  tavg_tilde_setup();
  time_setup();
  zunpack_setup();
  unpack_tilde_setup();
  urn_setup();
  wrap_setup();
  z_tilde_setup();
}
